﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;

namespace lunarLander
{
    /// <summary>
    /// Game kernel
    /// </summary>
    class Game
    {
        public Game()
        {
            bmp = new Bitmap(640, 480);
            statsFont = new Font(FontFamily.GenericMonospace, 9, GraphicsUnit.Pixel);
            Ground = new Ground();
            Ground.Width = 640;
            Ground.TotalHeight = 400;
            Ground.GenerateTerrain();
            Lander = new Lander();
            Lander.Game = this;
            Paused = false;
            keys = new HashSet<GameKeys>();
            whitePen = Pens.White;
            landingPen = new Pen(Brushes.White, 3);
            //load lander
            Lander.Sprite = (Bitmap)Bitmap.FromStream(System.Reflection.Assembly.GetExecutingAssembly().GetManifestResourceStream("lunarLander.lander2.png"));
        }

        #region public properties
        public Ground Ground { get; set; }
        public Lander Lander { get; set; }
        #endregion

        #region fields
        private PauseReason Reason;
        Font statsFont;
        Bitmap bmp;
        Pen whitePen;
        /// <summary>
        /// Pen for landing pad
        /// </summary>
        Pen landingPen;
        HashSet<GameKeys> keys;
        private bool Paused { get; set; }
        //Bitmap landerBmp;
        #endregion

        #region events
        /// <summary>
        /// Called on key down
        /// </summary>
        /// <param name="e"></param>
        public void KeyDown(KeyEventArgs e)
        {
            //KeysPressed |= GetKeyFromKeyEventArgs(e);
            keys.Add(GetKeyFromKeyEventArgs(e));
        }

        /// <summary>
        /// Called on key up
        /// </summary>
        /// <param name="e"></param>
        public void KeyUp(KeyEventArgs e)
        {
            //KeysPressed &= ~GetKeyFromKeyEventArgs(e);
            keys.Remove(GetKeyFromKeyEventArgs(e));
        }

        private GameKeys GetKeyFromKeyEventArgs(KeyEventArgs e)
        {
            GameKeys key;
            switch (e.KeyData)
            {
                case Keys.Up:
                    key = GameKeys.Up;
                    break;
                case Keys.Down:
                    key = GameKeys.Down;
                    break;
                case Keys.Left:
                    key = GameKeys.Left;
                    break;
                case Keys.Right:
                    key = GameKeys.Right;
                    break;
                case Keys.Space:
                    key = GameKeys.Space;
                    break;
                case Keys.X:
                    key = GameKeys.X;
                    break;
                default:
                    key = GameKeys.Unknown;
                    break;
            }
            return key;
        }

        /// <summary>
        /// Checks if left key is pressed
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public bool IsKeyDown(GameKeys key)
        {
            //return ((KeysPressed & key) == key);
            return keys.Contains(key);
        }
        #endregion

        public void Update(TimeSpan ts)
        {
            //check pause key
            if (IsKeyDown(GameKeys.Space))
            {
                if (Reason == PauseReason.Paused)
                    Paused = !Paused;
            }
            //update all game objects
            if (!Paused)
                Lander.Update(ts);
            //check for collisions
            for (int i = 0; i < Ground.points.Count - 1; i++)
            {
                if (Lander.IntersectsWithLine(Ground.points[i], Ground.points[i + 1]))
                {
                    //check if we crashed 
                    Paused = true;
                    Reason = PauseReason.Crashed;

                    if (Ground.points[i].Y == Ground.points[i + 1].Y)
                        if ((Lander.sY < Lander.MaxLandingSpeed) && (Lander.IsRotatedForLanding))
                            Reason = PauseReason.Landed;
                }
            }
        }

        public Bitmap Draw()
        {
            using (Graphics g = Graphics.FromImage(bmp))
            {
                g.Clear(Color.Black);
                //draw the ground
                for (int i = 0; i < Ground.points.Count - 1; i++)
                    g.DrawLine((Ground.points[i].Y == Ground.points[i + 1].Y) ? landingPen : whitePen, Ground.points[i], Ground.points[i + 1]);
                //draw the lander
                //transform graphics object by the rotation
                Matrix m = new Matrix();
                m.RotateAt((float)MathFuncs.RadiansToDegrees(Lander.Rotation), Lander.Location);
                g.Transform = m;
                g.DrawImage(Lander.Sprite, Lander.LocationGraphics);
                //draw fire if accelerating
                if (IsKeyDown(GameKeys.Up) || IsKeyDown(GameKeys.X))
                    g.DrawArc(Pens.White, Lander.Location.X - 3, Lander.Location.Y + 5, 6, 10, 180, 180);
                //restore transformation
                m.Reset();
                g.Transform = m;

                //draw stats
                const int posStatX = 0;
                g.DrawString(DateTime.Now.ToLongTimeString(), statsFont, Brushes.White, 590, 0);
                g.DrawString("Fuel: " + Lander.Fuel, statsFont, Brushes.White, posStatX, 0);
                g.DrawString(string.Format("Lander: x={0:0.00} y={1:0.00} rot={2:0.00}", Lander.X, Lander.Y, MathFuncs.RadiansToDegrees(Lander.Rotation)), statsFont, Brushes.White, posStatX, 10);
                g.DrawString(string.Format("Speed: x={0:0.00} y={1:0.00} rot={2:0.00}", Lander.sX, Lander.sY, Lander.RotationMomentum), statsFont, Brushes.White, posStatX, 20);

                if (Paused) //or game over
                    g.DrawString(Reason.ToString(), new Font(FontFamily.GenericMonospace, 80, FontStyle.Bold), Brushes.White, 50, 150);
            }
            return bmp;
        }


        enum PauseReason
        {
            Paused, Crashed, Landed
        }
    }
}