﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
namespace lunarLander
{
    class Lander
    {
        public Lander()
        {
            Fuel = 1000;
            X = 200;
            Y = 50;
        }

        public const float MaxLandingSpeed = 15;
        const float size = 12;
        double Gravity = 9.8; //1.7;
        double ThrustSpeed = 20;
        const double RotationBurnedFuel = 0.5;
        public Bitmap Sprite { get; set; }

        public Game Game { get; set; }

        /// <summary>
        /// X speed in m/s 
        /// </summary>
        public double sX { get; set; }

        /// <summary>
        /// Y speed in m/s
        /// </summary>
        public double sY { get; set; }

        /// <summary>
        /// Location of lander
        /// </summary>
        public PointF Location
        {
            get
            {
                return new PointF((float)X, (float)Y);
            }
        }

        /// <summary>
        /// Location of "top left" point of our graphics sprite
        /// </summary>
        public PointF LocationGraphics
        {
            get
            {
                return new PointF((float)X - size / 2, (float)Y - size / 2);
            }
        }

        /// <summary>
        /// rectangle of the sprite
        /// </summary>
        public RectangleF Rectangle
        {
            get
            {
                return new RectangleF((float)X - size / 2, (float)Y - size / 2, size, size);
            }
        }

        /// <summary>
        /// X location
        /// </summary>
        public double X { get; set; }

        /// <summary>
        /// Y location
        /// </summary>
        public double Y { get; set; }

        /// <summary>
        /// current angle in rads
        /// </summary>
        public double Rotation { get; set; }

        /// <summary>
        /// current angle in degrees
        /// </summary>
        public double RotationDegrees
        {
            get { return Rotation / Math.PI * 180; }
        }

        /// <summary>
        /// Updates langer with specified timespan
        /// </summary>
        /// <param name="ts"></param>
        public void Update(TimeSpan ts)
        {
            //update location according to time passed and speed
            System.Diagnostics.Trace.WriteLine("Grav:" + Gravity + " seconds:" + ts.TotalSeconds);
            
            //apply forces

            //gravity
            sY += Gravity * ts.TotalSeconds;

            //thrust
            if (Game.IsKeyDown(GameKeys.Up))
                Thrust(ts.TotalSeconds);
            if (Game.IsKeyDown(GameKeys.X))
                Afterburner(ts.TotalSeconds);
            if (Game.IsKeyDown(GameKeys.Left))
                RotateLeft(ts.TotalSeconds);
            if (Game.IsKeyDown(GameKeys.Right))
                RotateRight(ts.TotalSeconds);

            //calculate new location
            X += sX * ts.TotalSeconds;
            Y += sY * ts.TotalSeconds;

            //calculate new rotation
            Rotation += RotationMomentum * ts.TotalSeconds;
            //we don't lose rotation momentum as we are in the space, duh!
            Rotation = MathFuncs.NormalizeAngle(Rotation);
        }

        public double RotationMomentum { get; set; }

        const float rotationCoeff = 2.0f;
        /// <summary>
        /// Apply rotation throttle
        /// </summary>
        private void RotateLeft(double totalSeconds)
        {
            if (Fuel < 0)
                return;
            RotationMomentum -= Math.PI / rotationCoeff * totalSeconds;
  
            //decrease fuel
            Fuel -= RotationBurnedFuel;
        }

        private void RotateRight(double totalSeconds)
        {
            if (Fuel < 0)
                return;
            RotationMomentum += Math.PI / rotationCoeff * totalSeconds;
           
            //decrease fuel
            Fuel -= RotationBurnedFuel;
        }


        /// <summary>
        /// Thrust the engines
        /// </summary>
        private void Thrust(double totalSeconds)
        {
            if (Fuel < 0)
                return;
            //lander rotation 
            //add thrust
            sY -= Math.Sin(MathFuncs.NormalizeAngle(Rotation + Math.PI / 2)) * totalSeconds * ThrustSpeed;
            sX -= Math.Cos(MathFuncs.NormalizeAngle(Rotation + Math.PI / 2)) * totalSeconds * ThrustSpeed;
            //decrease fuel
            Fuel--;
        }

        /// <summary>
        /// Thrust the afterburner engine
        /// </summary>
        private void Afterburner(double totalSeconds)
        {
            if (Fuel < 0)
                return;
            //add more thrust
            sY -= Math.Sin(MathFuncs.NormalizeAngle(Rotation + Math.PI / 2)) * totalSeconds * ThrustSpeed * 5;
            sX -= Math.Cos(MathFuncs.NormalizeAngle(Rotation + Math.PI / 2)) * totalSeconds * ThrustSpeed * 5;

            //decrease more fuel
            Fuel -= 10;
        }

        public double Fuel { get; set; }

        public bool IntersectsWithLine(PointF a, PointF b)
        {
            //get all 4 points of our bounding square and rotate them around its center
            //top left
            if (TestPointUnderLine(a, b, new PointF(Rectangle.Left, Rectangle.Top)))
                return true;

            //top right
            if (TestPointUnderLine(a, b, new PointF(Rectangle.Right, Rectangle.Top)))
                return true;

            //bottom left
            if(TestPointUnderLine(a,b,new PointF(Rectangle.Left,Rectangle.Bottom)))
                return true;

            //bottom right
            if (TestPointUnderLine(a, b, new PointF(Rectangle.Right, Rectangle.Bottom)))
                return true;


            return false ;
        }

        /// <summary>
        /// Checks if point is "under" a line
        /// </summary>x
        /// <param name="left"></param>
        /// <param name="right"></param>
        /// <param name="point"></param>
        /// <returns></returns>
        private bool TestPointUnderLine(PointF left, PointF right, PointF point)
        {
            //skip if point is not in x-range of our line
            if (point.X < left.X || point.X > right.X)
                return false;

            //basically test if Line.GetYCoordAtX(point.X) > point.Y
            //get elements for line equation
            //we want to get coords of each point on line by using y=a*x+b

            //a is increment for each X
            float a = (right.Y - left.Y) / (right.X - left.X);
            //b is the starting y
            float b = left.Y;

            //we make test as if point started at x=0 so we shift its X location relative to start of the line
            //and store it in newX
            float newX = point.X - left.X;

            //if point is "under" the line, we 
            return (point.Y > (a * newX + b));
        }

        public bool IsRotatedForLanding
        {
            get
            {
                return ((RotationDegrees > 350) || (RotationDegrees < 10));
            }
        }
    }
}
